#ifndef MODULES_DPC_UTILS_INCLUDE_SKELETONWRAPPER_H_
#define MODULES_DPC_UTILS_INCLUDE_SKELETONWRAPPER_H_

#include "ara/com/internal/skeleton/skeleton_adapter.h"
#include "ara/com/internal/skeleton/event_adapter.h"
#include "ara/com/internal/skeleton/field_adapter.h"
#include "ara/com/internal/skeleton/method_adapter.h"
#include "ara/log/logging.h"

#include <string>

#include "PubDataWrapper.h"

#include <type_traits>

template <typename T, typename = int>
struct HasX : std::false_type { };

template <typename T>
struct HasX <T, decltype((void) T::data, 0)> : std::true_type { };

template <typename PublisherType, typename EventType, typename DataType>
class PubWrapper
{
private:
	std::string _name;
	PubDataWrapper<DataType>& _dataWrapper;
	ara::log::Logger& _log;
	bool _isInitialized;

	std::vector<ara::com::SampleAllocateePtr<DataType>> _memPool;
	const int max_mem_pool_size = 10;
	bool _refill;
	std::thread _thread;
	std::vector<int> _memStatus;


	virtual EventType& GetEvent() = 0;

	ara::com::SampleAllocateePtr<DataType> Allocate()
	{
		return GetEvent().Allocate();
	}

	void Loop()
	{
		while (1)
		{
			for (auto i = 0; i < max_mem_pool_size; ++i)
			{
				if(_memStatus[i] == 2)
				{
					_memPool[i] = Allocate();
					if constexpr(HasX<DataType>::value == true)
						_memPool[i]->data.reserve(14 * 150000);
					_memStatus[i] = 0;
				}
			}
			std::this_thread::sleep_for(std::chrono::milliseconds(10));
		}
	}

public:
	PubWrapper<PublisherType, EventType, DataType>(PubDataWrapper<DataType>& dataWrapper, const std::string& name, ara::log::Logger& log) : _dataWrapper(dataWrapper), _name(name), _log(log)
	{
		_refill = false;
		_isInitialized = false;
		_log.LogInfo() << "Skeleton for: " << _name << " created";
	}

	virtual ~PubWrapper<PublisherType, EventType, DataType>(){}

	std::shared_ptr<PublisherType> _skeleton;

	void Init(bool use_mempool = false)
	{
		_log.LogInfo() << "Initializing skeleton for: " << _name;
		_skeleton = std::make_shared<PublisherType>(ara::core::InstanceSpecifier(_name.c_str()),
				ara::com::MethodCallProcessingMode::kEvent);
		_skeleton->OfferService();

		if(use_mempool)
		{
			for (auto i = 0; i < max_mem_pool_size; ++i)
			{
				_memPool.push_back(Allocate());
				_memStatus.push_back(0);
			}

			_thread = std::thread(&PubWrapper<PublisherType, EventType, DataType>::Loop, this);
		}

		_isInitialized = true;
		_log.LogInfo() << "Skeleton for: " << _name << " initialized";
	}

	void Init(std::shared_ptr<PublisherType> skeleton, bool use_mempool = false)
	{
		_skeleton = skeleton;
		_isInitialized = true;
		if(use_mempool)
		{
			for (auto i = 0; i < max_mem_pool_size; ++i)
			{
				_memPool.push_back(Allocate());
				_memStatus.push_back(0);
			}

			_thread = std::thread(&PubWrapper<PublisherType, EventType, DataType>::Loop, this);
		}
		_log.LogInfo() << "Skeleton for: " << _name << " initialized";
	}

	int AllocateInternal()
	{
		for(auto i = 0 ; i < _memStatus.size() ; ++i)
		{
			if(_memStatus[i] == 0)
				return i;
		}

		return -1;
	}

	ara::com::SampleAllocateePtr<DataType>& GetDataPtr(int i)
	{
		_memStatus[i] = 1;
		return _memPool[i];
	}


	void Publish()
	{
		if(!_isInitialized)
		{
			_log.LogError() << "Skeleton must be initialized before it is used @ " << _name ;
			return;
		}
		auto data = Allocate();
		_dataWrapper.GetData(data);
		GetEvent().Send(std::move(data));
	}

	void Publish(ara::com::SampleAllocateePtr<DataType> data)
	{
		if (!_isInitialized)
		{
			_log.LogError()
					<< "Skeleton must be initialized before it is used @ "
					<< _name;
			return;
		}
		GetEvent().Send(std::move(data));
	}

	void PublishRaw(int i)
	{
		if (!_isInitialized)
		{
			_log.LogError()
					<< "Skeleton must be initialized before it is used @ "
					<< _name;
			return;
		}

		GetEvent().Send(*_memPool[i].get());
		_memStatus[i] = 2;
	}
};


#endif /* MODULES_DPC_UTILS_INCLUDE_SKELETONWRAPPER_H_ */
